#! /usr/local/bin/perl -- # -*-Perl-*-

########################################################################
# TeXit: a Perl script for running TeX.  
#
# Copyright (C) 1993 by Norman Walsh.  <walsh@cs.umass.edu>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Comments, questions, suggestions, and even flames welcome.
#
# See "display_usage" at the bottom of this file for information
# about how to use TeXit.
########################################################################

($VERSION = "\$Revision: 1.24 $") =~ s/.*: (\d+.\d+).*/$1/;
printf "This is TeXit v$VERSION.  ";
printf "Copyright (C) 1993 by Norman Walsh.\n";

########################################################################
# Load standard library routines for handling abbreviations.  I use
# this for the command line options.
#
require "abbrev.pl";

########################################################################
# Initialize associative array of format filenames.  
#
#   $formats{common-name} = name-of-format-file
#
$formats{"amslatex"} = "lplain";
$formats{"amstex"}   = "amstex";
$formats{"foiltex"}  = "fltplain";
$formats{"lamstex"}  = "lamstex";
# $formats{"latex"}    = "lplain";
$formats{"latex"}    = "latex";
$formats{"lollipop"} = "lollipop";
$formats{"plain"}    = "plain";
$formats{"slitex"}   = "splain";
$formats{"texinfo"}  = "plain";

########################################################################
# Global defaults.  These should be set to reasonable values when the 
# script is installed.  The values below are reasonable for many Unix
# installations. 
#
# When TeXit begins, it loads each of the following files, if they
# exist:
#
#   A system-wide init file (if configured, this file _must_ exist)
#   ~/.texitrc 
#   /path/.texitrc (/path is the path of the tex file being processed)
#   foo (for each -initfile 'foo' specified on the command line)
#
# They are loaded in that order and each can over-ride the values of
# the preceding.  Or any of the defaults shown below:
#
$DEFAULTFORMAT = "latex";     # Default format to use...
$VERBOSE       = 0;           # Lots of messages...
$TEXITFASTEXIT = 1;           # Empty line at prompt == quit
$ONCE          = 0;           # Run once, no prompting?
$CONFIRMPRINT  = 1;           # Ask before printing?

$QuitReply     = "e[x]it";    # To leave TeXit
$TeXReply      = "[t]ex";     # To run TeX
$CleanupReply  = "[c]leanup"; # To cleanup files

$TEXITRC       = ".texitrc";  # Filename for init files ~/.texitrc ...
$TEXRETRIES    = 4;           # How many retries before we give up?
$TEXHEADERSIZE = 20;          # number of lines in header material

$TEX           = "virtex";    # TeX, the program
$BIBTEX        = "bibtex";    # BibTeX
$MAKEINDEX     = "makeindex"; # MakeIndex
$VIEW          = "xdvi";      # A previewer
$GSVIEW        = "ghostview"; # Ghostview, another previewer

$TEXOPTS       = "";          # Default options for each of the above
$BIBTEXOPTS    = "";          #   These options are _always_ used if
$MAKEINDEXOPTS = "";          #   they are set, so choose wisely.
$VIEWOPTS      = "";
$GSVIEWOPTS    = "";

# The following commands are used to run each of the above programs.
# See the 'run' subroutine for a description of the %-variable 
# substitution that is performed.  Also, note that Perl $VARIABLE
# substitution is performed too.
#
$TEXCMD        = "$TEX \"%F %o \\input{%s}\"";
$BIBTEXCMD     = "$BIBTEX %o %s";
$MAKEINDEXCMD  = "$MAKEINDEX %o %s";
$VIEWCMD       = "$VIEW %o %d";
$GSVIEWCMD     = "$GSVIEW %o %s.ps";

########################################################################
# Printer initialization.  It works like this: the user selects a 
# printer queue "Q".  Then $PRQUEUE{"Q"} determines the printer _type_.
# Ordinarily, the user never knows about the type, but it is used
# to lookup the DVI driver and printer commands.  In particular:
#
# $PRQUEUE{"Q"}   = "TYPE"
#
# $DVICMD{"TYPE"} = DVI driver command (with substitutions)
# $DVIOPT{"TYPE"} = Default options for driver
# 
# $PRINTCMD{"Q"}  = How to print resulting file (with substitution)
# $PRINTOPT{"Q"}  = Default options for driver
#
# Ordinarily, all of the above variables should be set for each queue
# and it is an error for them to be unset.  However, the following
# exception is made: if the $PRINTCMD{"Q"} isn't set, then the file
# is run through a DVI driver but is not printed.
#
# Note: Ghostview is implemented this way, but it is also hard coded
# later on as an option in it's own right because we can be a little
# more intelligent about it.
#
# It would clearly have been possible to do configuration in other
# ways, in particular by using a single print command with different
# printer options ("lpr -P%p" for example).  But that has disadvantages
# too and it doesn't save much time or space.
#
$PRQUEUE{"Local"}     = "ps";    # A 'local' printer
$PRQUEUE{"File"}      = "ps";    # Just save the PS file
$PRQUEUE{"ghostview"} = "ps";    # Ghostview

$PRINTCMD{"Local"} = "lpr %o %s.ps";
$PRINTOPT{"Local"} = "";

$PRINTCMD{"File"}  = "";
$PRINTOPT{"File"}  = "";

$PRINTCMD{"ghostview"} = "$GSVIEW %o %s.ps";
$PRINTOPT{"ghostview"} = "";

$DVICMD{"ps"} = "dvips %o %d";
$DVIOPT{"ps"} = "";

########################################################################
# The $CleanupReply option looks for a list of files to delete in the 
# list "@cleanup_files".  These are usually set by perl commands in the 
# document since there aren't any default _files_ to delete.
#
# It also looks for a list of extensions in the list "@cleanup_<format>"
# If the file <texbasefile>.<ext> exists, it is deleted during cleanup
#
@cleanup_files = ();
@cleanup_plain = ('dvi', 'log');
@cleanup_latex = ('dvi', 'aux', 'log', 'blg', 'ilg');

########################################################################
# The system default, user default, and directory default initfiles
# have to be kept around until we're ready to load them.  That's what
# the @TEXITINIT list is for.  Ordinarily this is initialized empty.
# But if you always want to load a system-wide init file, perhaps
# "/usr/local/lib/texitrc" then you can initialize TEXITINIT to 
# include that file.
#
@TEXITINIT = ();

# Unfortunately, we want to run the system default files before the
# files specified with -initfile on the command line, but we don't
# find out what the name of the directory default is until after we've
# parsed the command line.  To circumvent this problem, we store
# initfiles from the command line in a seperate array.
#
@TEXITCMDLINE = (); # leave this empty

########################################################################
# Locate the user-default init file and store it on the TEXITINIT list.
# Some installations might not have a HOME environment variable to
# contain the home directory, so we let them specify a TEXIT env. var.
# instead.
#
$_ = $ENV{"TEXIT"}; 
$_ = $ENV{"HOME"} if $_ eq "";
$_ .= "/"         if $_ && $_ !~ /.*\/$/;
$texitrc = $_ . $TEXITRC;
printf "Looking for user-init file: $texitrc\n" if $VERBOSE;
unshift(@TEXITINIT, $texitrc) if -r $texitrc;

########################################################################
# Get ready to parse the arguments...
#
# The associative array "cmdlineopts" is built to hold all of the 
# possible, unambiguous command line options.  It will later be used
# to get back the full name (i.e. $cmdlineopts{"ini"} = "initfile").
#
{
  local (@formatlist) = keys (%formats);
  %cmdlineopts = ();
  &abbrev (*cmdlineopts, 
           'initfile', 'verbose', 'once',
	   'bibtex', 'makeindex', 'tex',
	   @formatlist);
}

# Build the case statement for parsing the options.  This code builds
# an associative array and 'eval's it to get the cases.  It's a little
# obscure, but it works.
#
# If a valid option is not found in the case statement, it is assumed
# to be the name of a format file.  This saves a whole bunch of cases.
# Note: the lines inbetween &case and ENDCASE _are not_ evaluated
# by the Perl compiler at compile time...they are 'eval'ed at runtime.
#
&case (*casefoo,<<'ENDCASE');
  initfile   $texitrc = $ARGV[0] || "no file"; shift; 
  verbose    $VERBOSE = 1;
  once       $ONCE = 1;
  tex        $runtexfirst = 1; 
  bibtex     $runbibtexfirst = 1;
  makeindex  $runmakeindexfirst = 1;
ENDCASE

# make sure a few things are undefined
undef ($TEXFORMAT);
undef ($texitrc);
undef ($runtexfirst);
undef ($runbibtexfirst);
undef ($runmakeindexfirst); 

# Now look at the arguments, keep scanning through them as long as
# they begin with a hyphen.  Anything after that has to be a filename
#
while ($ARGV[0] =~ /^-(.*)/ && shift) {
  ($option = $1) =~ tr/A-Z/a-z/;
  $option = $cmdlineopts{$option}; # find long opt
  die "Unrecognized or ambiguous option: -$1.\n"
    if $option eq "";

  if (defined($casefoo{$option})) {
    # Ok, it's one of the arms of the case statement
    eval $casefoo{$option};
    if (defined($texitrc)) {
      # The only arm that isn't self-contained is the -initfile arm.
      # For that option, we have to check for the initfile and save
      # it for later.
      die "You forgot the file name after \"-$1\".\n" 
        if $texitrc eq "no file";
 
      die "Initfile does not exist: $texitrc\n" 
        if ! -r $texitrc;

      printf "Looking for -initfile: $texitrc\n" if $VERBOSE;
      unshift(@TEXITCMDLINE, $texitrc);
      undef ($texitrc);
    }
  } else {
    # It's not one of the arms of the case, so it must be a format.
    $TEXFORMAT = $option;
  }
}

########################################################################
# What's left on the command line must be the name of the file that
# contains the document to process.  Anything else must be options to
# TeX.  Make sure they begin with an escape (you can put \relax
# in front if you need to).  Note: this verifies that the user 
# escaped the shell command line correctly.
#
$texfile = $ARGV[0] || &display_usage(); # display_usage doesn't return
shift;

if (defined($ARGV[0])) {
  local ($texescape, $printescape);
  $texescape = "\\\\";
  $texescape = "@" if $TEXFORMAT eq "texinfo";
  eval "\$printescape = \"$texescape\"";
  if ($ARGV[0] !~ /^$texescape.*/) {
    printf "TeX options must begin with a \"$printescape\".  ";
    die "Use ${printescape}relax if necessary.\n";
  }

  $TEXOPTS .= " " . join (" ", @ARGV);
}

$texfile = &cleanup_texfilename($texfile);
die "Cannot read TeX file: $texfile\n" if (! -r $texfile);

########################################################################
# Maybe the file has an emacs-style "local variables" definition that
# describes the file.  Here, we only care about the local variables
# "TeX-master" and "mode".  These are the same variables AUC-TeX uses.
# The TeX-master file means that _this_ file is some sub-part of a
# larger document.  The larger document is pointed to by TeX-master.
# The "mode" is what format this is in (Plain TeX, LaTeX, ... etc.)
#
# Alternatively, if you don't use Emacs, you can get the same logic
# out of TeXit by putting:
#
# % Master: file 
# % Mode: format % or Format: format
# 
# somewhere near the top (within the first $TEXHEADERSIZE lines) of
# your document.  In this case, the line must begin with a percent
# sign and there must be nothing else on the line.  The Master: and
# Mode: commands must each appear on a line of their own.  If you
# specify both a Master: and a TeX-master in the local variables, the
# TeX-master will win (simply because it comes later in the file).
# The same goes for mode:.
#
# Other than that, TeXit tries to be smart about what format to use, 
# looking for things that are clues to the format.  Unfortunately, 
# TeXit is easy to fool, so you may need a -<format> on the command 
# line sometimes.
#
&get_local_vars ($texfile);

# If this file has a master, load the local vars from the master file
# as well...these override variables from the initial file, but that's
# probably for the best anyway...
#
# TeX-master = "t" is special, it means _this_ file is the master file
# (i.e. it's `true' that this is the master file).
#
# If there isn't a master file, then $texfile is the master.
#
$texmaster = $texvars{"TeX-master"};
if (defined($texmaster) && $texmaster ne "t") {
  $texmaster = &cleanup_texfilename($texmaster);
  printf "Processing $texfile ... superceded by master file.\n";
  die "Cannot read master file: $texmaster" if (! -r $texmaster);
  &get_local_vars ($texmaster);
} else {
  $texmaster = $texfile;
}

########################################################################
# Split the master filename up into a path, name, and extension
#
($TEXFILEPATH,$TEXFILENAME,$TEXFILEEXT) = &splitfn($texmaster);

########################################################################
# Build up the default array of commands.
#
# Prompt with "key" if $CMDPROMPT{key} exists and the test $CMDTEST{key} 
# succeeds.  A null test succeeds by default.  The string enclosed in 
# square brackets in the key is what the user has to enter to select 
# that option.
#
# TeXit `eval's $CMDPROMPT{key} if the user selects the option "key".
#
# Use single quotes when building the command to be evaluated _and_ 
# the test so that $VAR substitution isn't performed prematurely!
#
# The commands to run tex, cleanup the directory, and quit are used
# elsewhere.  Change the definitions of $xxxREPLY to make them different.
# But don't change what they _mean_!!!
#

$RETEX = "last ASKLOOP";
$REASK = "next ASKLOOP";

$CMDPROMPT{$TeXReply}     = '$RETEX';
$CMDPROMPT{$CleanupReply} = '&cleanup("$opts"); $REASK';
$CMDPROMPT{$QuitReply}    = 'exit(0)';
$CMDTEST{$TeXReply}       = '';
$CMDTEST{$CleanupReply}   = '&files_to_cleanup()';
$CMDTEST{$QuitReply}      = '';

# It would be ok to change these...
#
$CMDPROMPT{"[b]ibtex"}    = '&bibtexcmd("$opts"); $RETEX';
$CMDPROMPT{"[g]hostview"} = '&ghostviewcmd("$opts"); $REASK';
$CMDPROMPT{"[p]rint"}     = '&printcmd("$opts"); $REASK';
$CMDPROMPT{"[v]iew"}      = '&viewcmd("$opts"); $REASK';

$CMDTEST{"[b]ibtex"}      = '"$BIBTEX"  && -r &auxfile("aux")';
$CMDTEST{"[g]hostview"}   = '"$GSVIEW"  && -r &auxfile("dvi")';
$CMDTEST{"[p]rint"}       = '"%PRQUEUE" && -r &auxfile("dvi")';
$CMDTEST{"[v]iew"}        = '"$VIEW"    && -r &auxfile("dvi")';

########################################################################
# Look for a directory-local init file.  Then load all the init files
# that we've found.  Load the global one, then the user-global one,
# then the directory-local one then all the user specified ones, in
# order.
# 
# Note: init files are just bits of Perl code so they can really do
# anything you want.  But they have to be legal Perl.  Usually they
# just set some global vars.
#
# Yes, "require_hack" is ugly.  But read the comments in that routine
# and you'll see why it was necessary.
#
$texitrc = $TEXFILEPATH . $TEXITRC;
printf "Looking for local init file: $texitrc\n" if $VERBOSE;
unshift (@TEXITINIT, $texitrc) if -r $texitrc;

while (@TEXITINIT) {
  $texitrc = pop(@TEXITINIT);
  if (-r $texitrc) {
    &require_hack ($texitrc);
  } else {
    warn "Can't read initfile: $texitrc\n";
  }
}

while (@TEXITCMDLINE) {
  $texitrc = pop(@TEXITCMDLINE);
  if (-r $texitrc) {
    &require_hack ($texitrc);
  } else {
    warn "Can't read initfile: $texitrc\n";
  }
}

undef($texitrc);

########################################################################
# Now that all the initialization code has been run, build up an
# abbreviation list for the printer queues.
#
{ 
  local ($queue,@prlist);

  foreach $queue (keys %PRQUEUE) {
    unshift(@prlist, $queue);
  }

  %pr_queue_names = ();
  &abbrev (*pr_queue_names, @prlist);
}

########################################################################
# Pick a format.  Tell the user if we don't agree about the format
# they demanded.  Shift the format to lowercase for using as a key in 
# the %formats associative array.
#
# Bail out if we can't find the format in the %formats.
#
printf "Processing $TEXFILEPATH$TEXFILENAME.$TEXFILEEXT ... ";
if (defined($texvars{"mode"}) && defined($TEXFORMAT)
    && $TEXFORMAT ne $texvars{"mode"}) {
  printf "it looks like a $texvars{mode} file to me,\n";
  printf "but you said it was a $TEXFORMAT file.  ";
  printf "Suit yourself.\n";
  } elsif (!defined ($TEXFORMAT)) {
  printf "it looks like a $texvars{mode} file to me.\n";
  }

$TEXFORMAT = $TEXFORMAT || $texvars{"mode"};
$TEXFORMAT =~ tr/A-Z/a-z/;

die "Don't know what format file to use for \"$TEXFORMAT\"."
  if (!defined($formats{$TEXFORMAT}));

########################################################################
# Evaluate the PLC commands for this document. 
#
# PLC commands are bits of Perl that usually look something like this:
#
# %  &makeindex('-o driver.ind driver.idx') 
# %    if &missing('driver.ind') || &changed('driver.idx');
# %  &bibtex('driver') if &newer('jigsaw.bib','driver.bbl');
#
# This says run MakeIndex if "driver.ind" is missing or "driver.idx"
# changes over the course of a TeX run.  And run BibTeX if the 
# bibliography file is newer than the BBL file (it'll also get run 
# automatically if any citations are unresolved).
#
# $first_time is true now so that &missing, &dependson, &changed, and 
# &newer all return false regardless of the condition tested.  This
# way we can initialize some globals before we run TeX for the first 
# time (to see if files change on the first run, for example).
#
# VERBOSE is forced off temporarily so that no stupid messages are 
# printed, even if the user wants them.  Note the dynamic scoping 
# nature of Perl.
#

{ 
  local ($VERBOSE) = 0;
  local ($first_time) = 1;
  local ($line);

  for $line (@PLCRULES) {
      eval $line;
      die "Invalid PLC rule: $line\n" if $@;
  }
}

if ($VERBOSE && @PLCRULES) {
    printf "Your document defined the following rules:\n";
    for $line (@PLCRULES) {
	printf "  $line\n";
    }
    printf "\n";
}

########################################################################
# If the user asked us to run MakeIndex or BibTeX first, run them now.
#
if ($runbibtexfirst) {
  if (-r &auxfile("aux")) {
    &bibtexcmd("");
  } else {
    printf "There's no AUX file to run BibTeX against.\n";
  }
}
undef($runbibtexfirst);

if ($runmakeindexfirst) {
  if (-r &auxfile("idx")) {
    &makeindexcmd("");
  } else {
    printf "There's no IDX file to run MakeIndex against.\n";
  }
}
undef($runmakeindexfirst);

########################################################################
# Ok, now it's time to run TeX.  Run TeX repeatedly until an error 
# occurs, the dependencies don't change or we've tried more than four 
# times.  If it doesn't work in four tries, it never will.
#
$opts = "";     # user specified options after the prompt
$changed = 0;   # did the file change on this run?
$runcount = 0;  # how many times have we run TeX in a row?

# Think about BibTeX.  After the first run, we've collected all the
# citations we're going to find so there's no point in running it again.
# This prevents a missing citation from causing the BBL file to change
# every time and thus causing TeX to be re-run every time even when it's
# bound to fail.
#
$ranbibtex = 0; # have we run bibtex?
$changed = 1;   # assume that TeX needs to be run at least once...

# if the tex file and the dvi file and the log file all exist, maybe
# we don't actually have to run TeX first...maybe the user just wants
# to print or preview or something and they're used to typing "texit".
#
if (!$runtexfirst) { # if the user doesn't insist...
    $changed = &dependencies_changed () 
	if (-r $texmaster && -r &auxfile("dvi") && -r &auxfile("log"));
}
undef($runtexfirst);

TEXLOOP: for(;;) {		# don't stop till the user says quit.
    if ($changed) {
	$rc = &texcmd($opts);	# run TeX
	$changed = 1, next TEXLOOP 
	    if $runcount < $TEXRETRIES       # more tries?
		&& $rc == 0	             # Tex succeeded?
	        && &dependencies_changed (); # not done
    } else {
	$rc = 0;
    }

    warn "\nTeX failed.  All bets are off.\n\n" if $rc;

    if ($runcount >= $TEXRETRIES && !$rc) {
	printf "\nI've tried $TEXRETRIES times and it still ";
	printf "isn't right...I give up.\n\n";
	exit 1 if $ONCE;
    } else {
	exit 0 if $ONCE;
    }

    $changed = 1; # assume it'll change before we get back to top...

    $lastcmd = "";
  ASKLOOP: for (;;) {
      # User can change anything while we're sitting at the prompt, so
      # reset runcount and ranbibtex.
      $runcount   = 0; 
      $ranbibtex  = 0;

      ($cmd, $opts) = &get_command();

      if ($cmd eq "") {
	  $cmd = $QuitReply;
	  $cmd = "redo" if !$TEXITFASTEXIT;

	  # Quit or redo as planned if last command was to cleanup.
	  # Otherwise re-run TeX if it's necessary...
	  if ($lastcmd ne $CleanupReply) {
	      if (! -r &auxfile("dvi")) {
		  print "\n", &auxfile("dvi"), " disappeared.\n";
		  $cmd = $TeXReply;
	      } elsif (&mtime(&auxfile("dvi")) < &mtime($texmaster)) {
		  printf "\n$texmaster has changed.\n";
		  $cmd = $TeXReply;
	      }
	  }
      }
      
      next if $cmd eq "redo";

      $lastcmd = $cmd;

      # Ok, now it's time to 'eval' the command.  First, copy it and
      # protect any backslashes and quotations in it from premature
      # replacement.  Then eval it once to do $-var substitution.
      # Then eval it again to really do the work...
      #
      $thecmd = $CMDPROMPT{$cmd};
      $thecmd =~ s/\\/\\\\/g;	# protect \'s
      $thecmd =~ s/\"/\\"/g;    # protect "'s
      eval "\$thecmd = \"$thecmd\"";
      eval ( $thecmd );
      die "Invalid command defined for $cmd: $CMDPROMPT{$cmd}\n" if $@;
  }
}

########################################################################
# These subroutines execute the expected commands.  They are here, with 
# nice names, so that the PLC commands can use them and be easy to read.
# Internally, different routines are used so that %-var substitution is
# preformed.  Note that it _isn't_ preformed here. 
#
# &tex("foo") ==> $TEX foo
#
sub tex {
  local ($opts) = @_;

  &run($TEX . " " . $TEXOPTS . " " . $opts, "");
}

sub bibtex {
  local ($opts) = @_;

  &run($BIBTEX . " " . $BIBTEXOPTS . " " . $opts, "");
}

sub makeindex {
  local ($opts) = @_;

  &run($MAKEINDEX . " " . $MAKEINDEXOPTS . " " . $opts, "");
}

########################################################################
# These are the subroutines used internally where all we want to pass 
# in are the _additional_ options.
#
# &texcmd("foo") => $TEXCMD with $TEXOPTS and "foo" inserted as options
#                   and all other %-var substitution performed.
#
sub texcmd {
  local ($opts) = @_;
  local ($rc);

  &run ($TEXCMD, $TEXOPTS . " " . $opts);

  $rc = $? >> 8;
  $runcount++;
  $rc;
}

sub bibtexcmd {
  local ($opts) = @_;

  &run ($BIBTEXCMD, $BIBTEXOPTS . " " . $opts);

  $? >> 8;
}

sub makeindexcmd {
  local ($opts) = @_;

  &run($MAKEINDEXCMD, $MAKEINDEXOPTS . " " . $opts);
}

sub viewcmd {
  local ($opts) = @_;

  &run ($VIEWCMD, $VIEWOPTS . " " . $opts);

  $? >> 8;
}

sub printcmd {
  local ($opts) = @_;
  local ($queueopt, $dviuseropts, $prtuseropts);
  local ($rc, $type);
  local ($dvicmd, $dviopt, $prtcmd, $prtopt);
  local ($doit,$queue);

  # Make sure the user gave us a printer queue
  if ($opts =~ /^\s*(\S+)\s*(.*)$/) {
    $queueopt = $1;
    $opts  = $2;
  } else {
    printf "You must specify a printer queue (use \"p <printer>\").\n";
    return(0);
  }

  # Special case, if the queue is "?", list the queues we know about.
  if ($queueopt eq "?") {
    &print_long_list("\nThe following queues are available: ",
                     sort(keys(%PRQUEUE)));
    printf "\n";
    return (0);
  }

  $queue = $pr_queue_names{$queueopt};  

  if (!defined ($PRQUEUE{$queue})) {
    warn "There is no printer queue named \"$queueopt\".\n";
    return 0;
  }

  ($dviuseropts, $prtuseropts) = split (/\|/, $opts);

  $type   = $PRQUEUE{$queue};

  if (!defined ($DVICMD{$type})) {
    warn "There is no DVI driver for \"$type\" printers.\n";
    return 0;
  }

  $dvicmd = $DVICMD{$type};
  $dviopt = $DVIOPT{$type};

  $prtcmd = $PRINTCMD{$queue};
  $prtopt = $PRINTOPT{$queue};

  &run ($dvicmd, $dviopt . " " . $dviuseropts);
  $rc = $? >> 8;

  warn "DVI driver failed.\n" if $rc;

  $doit = ($prtcmd ne "");	# really print it...

  # Wait!  If the driver failed or there's a queue other than 
  # Ghostview selected, ask and make sure.
  #
  if ($rc || ($doit && ($queue ne "ghostview") && $CONFIRMPRINT)) {
      printf "Really print to $queue? ";
      $_ = scalar (<STDIN>); chop;
      $doit = 0 if !/^y/i;
  }

  if ($doit) {
    &run ($prtcmd, $prtopt . " " . $prtuseropts);
    $rc = $? >> 8;
  } else {
    printf "Print command aborted.\n" if ($prtcmd ne "");
  }

  print "\n";

  $rc;
}

########################################################################
# Ghostview is a little special.  The print commands always have to 
# run the DVI driver because they don't know what kind of file might
# be produced, but with Ghostview, we know that it'll be the master
# document name with '.ps' on the end.  So, if that exists and is
# newer than the DVI file, don't bother running the DVI driver again.
#
# Note also that Ghostview knows what type it is, so the type doesn't
# have to be in the printer queue.  This means you can remove the
# Ghostview printer from the queue list even if you don't have any
# other PS printers.
#
# Oh, also, we always run the DVI driver again if the user entered
# DVI driver options at the prompt...
#
sub ghostviewcmd {
  local ($opts) = @_;
  local ($queue, $dviuseropts, $prtuseropts);
  local ($rc, $type);
  local ($dvicmd, $dviopt, $prtcmd, $prtopt);
  local ($doit);
  local ($psfile) = &auxfile("ps");

  ($dviuseropts, $prtuseropts) = split (/\|/, $opts);

  $type = "ps";

  if (!defined ($DVICMD{$type})) {
    warn "There is no DVI driver for \"$type\" output.\n";
    return 0;
  }

  $dvicmd = $DVICMD{$type};
  $dviopt = $DVIOPT{$type};

  if ((! -r $psfile || (&mtime($psfile) < &mtime(&auxfile("dvi"))))
      || $dviuseropts ne "") {
    &run ($dvicmd, $dviopt . " " . $dviuseropts);
    $rc = $? >> 8;
  } else {
    printf "\nPostScript file already exists.\n";
    $rc = 0;
  }

  $doit = 1;
  if ($rc) {
    warn "DVI driver failed.\n";
    printf "Preview anyway? ";
    $_ = scalar (<STDIN>); chop;
    $doit = 1 if /^y/i;
  }

  if ($doit) {
    &run ($GSVIEWCMD, $GSVIEWOPTS . " " . $prtuseropts);
    $rc = $? >> 8;
  }

  $rc;
}

########################################################################
# Go out there and find all the files that have to be cleaned up. 
# Return them in a list.
#
sub files_to_cleanup {
  local ($opts) = @_;
  local (%thefiles);
  local ($file, @filelist, $ext, @extlist);

  @filelist = ();

  # throw them all in an associative array first to
  # avoid duplicates.
  foreach $file (@cleanup_files) {
    $file =~ s/\\/\//; # \ -> /
    $file = $TEXFILEPATH . $file if $file !~ /\//;
    $thefiles{$file} = 1;
  }

  # find the list of extensions in "@cleanup_<format>".
  eval "\@extlist = \@cleanup_$TEXFORMAT";
  if (@extlist) {
      local ($file);
      foreach $ext (@extlist) {
	  $file = &auxfile($ext);
	  $thefiles{$file} = 1;
      }
  }

  # now put them in a regular array
  foreach $file (keys %thefiles) {
    if (-e $file) {
      unshift(@filelist, $file);
    }
  }

  # return that array

  @filelist;
}

########################################################################
# Delete any files that have to be cleaned up...
#
sub cleanup {
  local ($opts) = @_;
  local (@filelist, $_);

  @filelist = &files_to_cleanup();

  # if there are any files to be deleted, ask the user to confirm
  # and then blow 'em away...
  if (@filelist) {
      &print_long_list ("\nDelete: ", sort @filelist);
      printf "\n";
      printf "Are you sure? ";
      $_ = scalar (<STDIN>); chop;
      
      if (/^y/i) {  
	  unlink(@filelist);
      } else {
	  printf "Cleanup aborted.\n";
      }
      
      print "\n";
  }
}

########################################################################
# The workhorse.  This routine takes a command string and an options 
# string.  The command string is subjected to the following 
# interpolation:
#
# 1) It is "eval"ed so that Perl variables in the
#    string are replaced with their values, i.e.
#    "$TEX foo" -> "virtex foo" if $TEX="virtex".
# 
# 2) %-substitution is performed.  The following 
#    substitutions are currently supported:
#
#      %F = format name (&lplain)
#      %s = base document name (myfile)
#      %t = tex document name (myfile.tex)
#      %d = dvi document name (myfile.dvi)
#      %o = the options passed into run
#
sub run {
  local ($cmd,$opts) = @_;
  local ($format)  = "&" . $formats{$TEXFORMAT};
  local ($texbase) = $TEXFILEPATH . $TEXFILENAME;
  local ($texfile) = &auxfile ("tex");
  local ($dvifile) = &auxfile ("dvi");

  # $cmd may be of the form "$TEX ..." and we want to
  # expand the definition of $TEX, etc.
  $cmd =~ s/\\/\\\\/g;       # protect \'s
  $cmd =~ s/\"/\\\"/g;       # protect "'s
  eval "\$cmd = \"$cmd\"";
  
  # Now substitute for `%' options...
  $cmd =~ s/\%F/$format/g;
  $cmd =~ s/\%s/$texbase/g;
  $cmd =~ s/\%t/$texfile/g;
  $cmd =~ s/\%d/$dvifile/g;
  $cmd =~ s/\%o/$opts/g;

  $cmd =~ s/\s+/ /g; # remove extranous whitespace, (display hack).

  printf "\n$cmd\n";
  system "$cmd";
  $?;
}

########################################################################
# Gets a command from the user by building a prompt up from the 
# %CMDPROMPT and %CMDTEST arrays.  A particular entry is only allowed
# if an eval of it's %CMDTEST returns true.
#
# &get_command() returns a two element list.  The first element is
# index into the %CMDPROMPT array, the second element is any additional
# arguments supplied by the user.
#
sub get_command {
    local (%commands, $_, $options);
    local ($promptstring, $promptprefix);
    local ($line, $thetest, $result);

    %commands = ();
    $promptstring = "";
    $promptprefix = "Again? ";
    $line = $promptprefix;

    # look for a stupid configuration error where a CMDTEST has
    # gotten misspelled.
    foreach $cmdname (keys %CMDTEST) {
	die "$cmdname is defined as a CMDTEST but not as a CMDPROMPT.\n"
	    if !defined($CMDPROMPT{$cmdname});
    }

    foreach $cmdname (sort keys %CMDPROMPT) {
	$thetest = $CMDTEST{$cmdname};
	$result  = 1;
	if ($thetest ne "") {
	    eval "\$result = $thetest";
	    die "Bad CMDTEST: $thetest\n" if $@;
	}

	if ($result) {
	    local ($cmdkey) = $cmdname;
	    $cmdkey =~ s/^.*\[(.*)\].*$/$1/;
	    die "No command string in square brackets: $cmdname\n" 
		if $cmdkey eq "";
	    die "Duplicate command: $cmdname = $commands{$cmdkey}\n" 
		if defined($commands{$cmdkey});
	    $commands{$cmdkey} = $cmdname;
	    if (length($line . $cmdname) < 60) {
		$line .= $cmdname . ", ";
	    } else {
		$promptstring .= $line . "\n";
		$line = " " x length($promptprefix) . $cmdname . ", ";
	    }
	}
    }
    
    $line =~ s/(.*),\s*$/$1/;	# remove the last ", "...
    $promptstring .= $line . ": ";
    
    for (;;) {
	($cmdkey, $options) = ("", "");
	print $promptstring;
	$_ = scalar(<STDIN>); chop;
	($cmdkey, $options) = ($1, $2) if /\s*(\w+)\s*(.*)/;
	last if $commands{$cmdkey} ne "" || $_ eq "";
	printf "\nI don't understand the command: $cmdkey\n\n";
    }

    ($commands{$cmdkey}, $options);
}

########################################################################
# Parse the TeX file looking for a local variables section.  Build 
# the associative array "%texvars" to hold the variables.
#
sub get_local_vars {
  local($texfile) = @_;
  local(@parms, $in_locals, $near_top, $in_plc, $plc_line);

  # always set a mode of some sort.  Either Plain or the $DEFAULTFORMAT
  $texvars{"mode"} = "Plain";
  $texvars{"mode"} = $DEFAULTFORMAT if (defined($DEFAULTFORMAT));

  open (TEXFILE, $texfile) || die "Can't read $texfile.\n";

  $in_locals =  0; # have we found the start of the local vars?
  $in_plc    =  0; # are we in the PLC commands?
  $near_top  = $TEXHEADERSIZE; # are we in the "header" area?
  
  @PLCRULES = ();  # throw away any rules we've already found

  while (<TEXFILE>) {
    chop;

    if ($near_top) {
      $texvars{"TeX-master"} = $1 if /\%+\s*master:\s*(\S+)\s*$/i;
      $texvars{"mode"}       = $1 if /\%+\s*format:\s*(\S+)\s*$/i;
      $texvars{"mode"}       = $1 if /\%+\s*mode:\s*(\S+)\s*$/i;

      $texvars{"mode"} = "latex"    if (/^\\documentstyle/);
      $texvars{"mode"} = "slitex"   if (/^\\documentstyle.*slides\}/);
      $texvars{"mode"} = "amstex"   if (/^\\document\s*$/);
      $texvars{"mode"} = "lollipop" if (/^\\Define\w+:\w+/);
      $texvars{"mode"} = "texinfo"  if (/^\@node/);
      $texvars{"mode"} = "texinfo"  if (/^\@ifinfo/);
      $texvars{"mode"} = "texinfo"  if (/input texinfo/);

      $near_top--;
      }

    $in_locals = 0 if $in_locals && /End:\s*$/;

    if ($in_locals) {
	local($_) = $_;
	s/^\W*//; # remove leading whitespace
	@parms = split(/[: \t\"\']+/);
	$var = $parms[0];
	shift(@parms);
	$val = join(" ",@parms);
	$texvars{$var} = $val;
    }

    # the [V] makes sure Emacs doesn't get confused when editing
    # texit.pl
    $in_locals = 1 if /Local [V]ariables:\s*$/;

    if ($in_plc && /End\s+TeXit\s+rules:\s*$/) {
	push (@PLCRULES, $plc_line) if $plc_line ne "";
	undef($plc_line);
	$in_plc = 0;
    }

    if ($in_plc) {
	local ($line) = $_;
	local ($lastline) = /.*;\s*$/;         # ends with a semicolon?
	$line =~ s/\s*\%*\s*(.*)/$1/;          # trim leading whitespace
        $line =~ s/(.*)([^\\])(\#)(.*)$/$1$2/; # trim trailing comments
        $line = "" if $line =~ /^\#/;          # _only_ a comment?
	$plc_line .= $line . " " if $line ne "";
	if ($lastline) {
	    push (@PLCRULES, $plc_line);
	    $plc_line = "";
	}
    }
    
    $in_plc = 1 if /Start\s+TeXit\s+rules:\s*$/;
  }

  warn "Unended local variables section?\n" if $in_locals;
  die  "Unended TeXit rules section!\n" if $in_plc;

  close (TEXFILE); 
}

########################################################################
# Cleanup the TeX filename.  Add the extension ".tex" if it doesn't 
# already have an extension.
#
sub cleanup_texfilename {
  local($texfile) = @_;
  local($path,$base,$ext) = &splitfn($texfile);

  $ext = "tex" if ($ext eq "");

  $path . $base . "." . $ext;
}

########################################################################
# Break a filename into it's path, basename, and extension components.  
# The path returned always ends with a slash.  "./" is returned if the
# file has no path.
#
sub splitfn {
  local ($filename) = @_;
  local ($path, $basename, $ext) = ("", "", "");

  $filename =~ tr/\\/\//;    # translate \ into /
  
  if ($filename =~ /\//) {
    ($path = $filename) =~ s/\/[^\/]*$//;
    ($basename = $filename) =~ s/.*\///;
    }
  else {
    $path = ".";
    $basename = $filename;
  }

  if ($basename =~ /\./) {
    ($ext = $basename) =~ s/.*\.//;
    $basename =~ s/\.[^.]*$//;
  }

  ($path . "/",$basename,$ext);
}

########################################################################
# Look through the log file for the LaTeX warnings.
#
sub parse_latex_log {
  local ($newbib, $newref, $badref, $badfont) = (0, 0, 0, 0);
  local ($font, $reference, $citation);
  local ($logfile) = &auxfile("log");

  if (-r $logfile) {
    open (LOGFILE, $logfile);
    while (<LOGFILE>) {
      chop;
      if (/LaTeX Warning: Citation/) {
        $newbib = 1;
	($citation = $_) =~ s/.*`(.*)'.*/$1/;
	$citelist{$citation} = 1;
      }

      if (/LaTeX Warning: Reference/) {
	($reference = $_) =~ s/.*\`(.*)\'.*/$1/;
	$reflist{$reference} = 1;
	$badref = 1;
      }

      if (/Warning: Font\/shape.*undefined/) {
        ($font = $_) =~ s/.*\`(\S+)\s*\'.*/$1/;
	$fontlist{$font} = 1;
	$badfont = 1;
      }

      $newref = 1 if /LaTeX Warning: Label\(s\) may have changed/;
    }
    close (LOGFILE);
  }

  ($newbib, $newref, $badref, $badfont);
}

########################################################################
# Decide if TeX needs to be run again.  How do we decide?  Well, first
# we look at the log file, if we know the format, then we run the
# PLC commands.  The LaTeX log file might indicate that there's an
# unknown citation, a new reference (meaning Label(s) may have changed)
# a bad reference, or a bad font.  The last two are purely informative.
#
# The PLC commands file should change files that it indicates the document
# &dependson.  TeXit will notice that the DVI file is older than some
# file that it depends on and re-run TeX.
#
sub dependencies_changed {
  local ($newbib, $newref, $badref, $badfont);
  local ($citation, $reference);

  $changed = 0;

  # make sure the TeX file hasn't changed...
  if (! -r &auxfile("dvi")) {
      print &auxfile("dvi"), " is missing.\n" if $VERBOSE;
      $changed = 1;
  } elsif (&mtime(&auxfile("dvi")) < &mtime($texmaster)) {
      print "$texmaster is newer than ", &auxfile("dvi"), ".\n" 
	  if $VERBOSE;
      $changed = 1;
  }

  $newbib = 0;
  $newref = 0;
  $badref = 0;

  printf "\n";

  undef (%citelist);
  undef (%reflist);
  undef (%fontlist);
  ($newbib, $newref, $badref, $badfont) = &parse_latex_log() 
    if $TEXFORMAT eq "latex" 
       || $TEXFORMAT eq "foiltex"
       || $TEXFORMAT eq "amslatex";

  if ($newbib) {
    &print_long_list ("Missing citations: ", keys(%citelist));
    if ($ranbibtex) {
      printf "Bibliography database(s) may be incomplete.\n\n";
    } else {
      printf "\n"
    }

    # give up after two tries...
    if ($ranbibtex < 2) {
      &bibtexcmd ("");
      printf "\n";
      $changed = 1;
    }

    $ranbibtex++;
  }

  if ($badref) {
    &print_long_list ("Missing references: ", 
                      keys(%reflist));
    printf "\n";
    }

  if ($badfont) {
    &print_long_list ("FYI: substituted fonts: ", 
                      keys(%fontlist));
    printf "\n";
    }

  $changed = $changed || $newref;

  {
      local ($line);
      foreach $line (@PLCRULES) {
	  eval $line;
	  die "Invalid PLC line: $line\n" if $@;
      }
  }

  if (!$changed) {
    if ($newbib || $badref) {	
      printf "\n" if $VERBOSE;
      printf "The document is still incorrect, ";
      printf "but TeXing again won't help.\n";
      printf "Some citations cannot be found.  " 
        if $newbib;
      printf "Some references cannot be found.  " 
        if $badref;
      printf "\n\n";
    } else {
      printf "As far as I can tell, ";
      printf "the document is up to date.\n\n";
    }
  } else {
    printf "\n";
  }

  $changed;
}

########################################################################
# Input: an extension, `ext'
# Output: the filename $TEXFILEPATH . $TEXFILENAME . `.ext'
#
sub auxfile {
    local($ext) = @_;
    local($dot);

    $dot = "";
    $dot = "." if $ext ne "" && $ext !~ /^\./;

    $TEXFILEPATH . $TEXFILENAME . $dot . $ext;
}
 
########################################################################
# Input: list of files
# Output: true if any file is newer than DVI file
#
sub dependson {
  local(@filelist) = @_;
  local($file, $rerun);

  $rerun = 0;
  if (-r &auxfile("dvi")) {
    foreach $file (@filelist) {
      if (-r $file) {
        if (&mtime(&auxfile("dvi")) < &mtime($file)) {
          printf "$file newer than DVI file?  Yes.\n" 
	    if $VERBOSE;
	  $rerun = 1;
	} else {
	  printf "$file newer than DVI file?  No.\n" 
	    if $VERBOSE;
        }
      } else {
        printf "$file newer than DVI file?  No (doesn't exist).\n"
	  if $VERBOSE;
      }
    }
  }

  $changed = ($changed || $rerun) && (!$first_time);
  $rerun && (!$first_time);
}

########################################################################
# Input: list of files
# Output: true if any file is missing
#
sub missing {
  local(@filelist) = @_;
  local($missing, $file);

  $missing = 0;

  foreach $file (@filelist) {
    if (defined($ifmissing{$file})) {
      if (! -r $file) {
        printf "Missing: $file?  Yes.\n" if $VERBOSE;
	$missing = 1;
      } else {
        printf "Missing: $file?  No.\n" if $VERBOSE;
      }
    } else {
      $ifmissing{$file} = 1;
    }
  }

  $missing && (!$first_time);
}

########################################################################
# Input: list of files
# Output: waits until they all exist (and are readable)
#
sub waitfor {
  local(@filelist) = @_;
  local(@missing, $file);

  @missing = ();

  foreach $file (@filelist) {
      push (@missing, $file) if (! -r $file);
  }

  if ($VERBOSE && @missing) {
      &print_long_list ("\nWaiting for: ",@filelist);
      printf("\n");
  }

  foreach $file (@missing) {
      while (! -r $file) {
	  sleep 1;
      }
  }
}

########################################################################
# Input: list of files
# Output: true if any file has changed.  A file has changed if its 
#         checksum is different from the preceding call.  Files that
#         are missing both times are currently the same.
#
sub changed {
  local(@filelist) = @_;
  local($changed, $savecheck, $file);

  $changed = 0;

  foreach $file (@filelist) {
    if (defined($ifchanged{$file})) {
      $savecheck = &checksum($file);
      if ($ifchanged{$file} != $savecheck) {
        printf "Changed: $file?  Yes.\n" if $VERBOSE;
        $ifchanged{$file} = $savecheck;
 	$changed = 1;
      } else {
        printf "Changed: $file?  No.\n" if $VERBOSE;
      }
    } else {
      $ifchanged{$file} = &checksum($file);
    }
  }

  $changed && (!$first_time);
}

########################################################################
# Calculate a checksum for a file.
#      
sub checksum {
  local($file) = @_;
  local($/, $checksum);

  undef $/;

  if (-r $file) { 
    open (CHECKSUM, $file);
    $checksum = unpack("%32C*", <CHECKSUM>);
    close (CHECKSUM);
  } else {
    $checksum = "none"; 
  }

  $checksum;
}

########################################################################
# Input: two lists of files.  The lists must be quoted strings so that 
#        they can be seperated easily.  The two strings are split on 
#        "/, /"  so they _can_ include multiple files.
# Output: true if _any_ of the first list of files is newer than _any_ 
#         of the second list of files _or_ any of the second list of
#	  files is missing.
#
sub newer {
  local ($newfile, $oldfile) = @_;
  local (@newfiles,@oldfiles);
  local ($file, $missing, $newer);

  @newfiles = split(/[, ]/, $newfile);
  @oldfiles = split(/[, ]/, $oldfile);

  die "Bad call to &newer().  No files in `new' list.\n" 
    if !@newfiles;
  die "Bad call to &newer().  No files in `old' list.\n" 
    if !@oldfiles;

  $newer = 0;

  # Find the oldest old file.  All must exist.
  undef $oldfile;
  foreach $file (@oldfiles) {
    $newer = 1, $missing = $file, last if ! -r $file;
    $oldfile = $file 
      if (! $oldfile || (&mtime($file) < &mtime($oldfile)));
  }

  # Find the newest new file.
  undef $newfile;
  foreach $file (@newfiles) {
    $newfile = $file 
      if (-r $file 
         && (! $newfile || (&mtime($file) > &mtime($newfile))));
  }

  warn "None of the `new' files exist: @newfiles\n" 
    if ! $newfile;

  $newer = $newer 
             || ($newfile && (&mtime($newfile) > &mtime($oldfile)));

  if ($newer) {
    if ($oldfile) {
      printf "$newfile newer than $oldfile?  Yes.\n" 
        if $VERBOSE;
    } else {
      printf "$missing doesn't exist, so $newfile is newer.\n" 
        if $VERBOSE;
    }
  } else {
    printf "$newfile newer than $oldfile?  No.\n" 
      if $VERBOSE;
  }

  $newer && (!$first_time);
}

########################################################################
# This routine prints a message followed by a potentially long list of
# items, seperated by spaces.  It will never allow "word wrap" to occur
# in the middle of a word.  There has to be a better way, using Perl's
# report generation to do this, but I haven't looked yet.
#
sub print_long_list {
  local ($message,@thelist) = @_;
  local ($line) = $message;
  local ($item, $displaystring) = ("", "");

  foreach $item (@thelist) {
      if (length($line . $item) < 73) {
	  $line .= $item . ", ";
      } else {
	  $displaystring .= $line . "\n";
	  $line = $item . ", ";
      }
  }
  
  $line =~ s/(.*),\s*$/$1/;	# remove the last ", "...
  $displaystring .= $line . "\n";

  print $displaystring;
}  

########################################################################
# Builds an associative array for a pascal-like case statement.  This 
# is a bit obscure and relies on Perl's ability to evaluate strings
# that look like code at runtime.
sub case {
  local(*assoc,$_) = @_;
  for (split(/\n/)) {
    /^\s*(\S+)\s+(.*)/;
    for (eval $1) {
      $assoc{$_} = $2;
    }
  }
}

########################################################################
# This hack gets around problems with MS-DOS and OS/2 ports of Perl.
# Some ports get confused when presented with "require d:/path/file"
# because they think "d:/path/file" should appear somewhere in @INC
# because it doesn't begin with a slash.  Sigh.  This code isn't as
# nice as "require" but it works...
#
sub require_hack {
  local($file) = @_;
  local($line);

  printf "Loading init file: $file\n" if $VERBOSE;

  $line = "";
  open (INFO, $file);
  while (<INFO>) {
    $line = $line . $_;
    if (/;\s*$/) {
      eval $line;
      die "Error in init file: $file at line $..\n" if $@;
      $line = "";
    }
  }
  close (INFO);

  die "Missing \";\" at end of init file: $file.\n" 
    if ($line !~ /^\s*$/);
}

########################################################################
# This hack gets around problems with (at least) the OS/2 port of Perl.
# The OS/2 port (maybe the MS-DOS port?) have one _hell_ of an annoying
# bug.  The test "-M file" returns a large positive number instead of
# a negative number for files that are newer than the start-time of the
# script (someone probably did an unsigned computation where a signed
# computation was required...)  The result is that -M file1 > -M file2
# if file2 existed when the script began and file1 did not.  Really
# bad.
#
# Caveat, on a working system:
#
#  -M file1 > -M file2 ==> &mtime(file1) < &mtime(file2)
#
sub mtime {
    local ($file) = @_;
    local ($dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size,
	   $atime, $mtime, $ctime, $blksize, $blocks);

    if (-r $file) {
	($dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size,
	 $atime, $mtime, $ctime, $blksize, $blocks) = stat($file);
    } else {
	$mtime = -1;
    }

    $mtime;
}

########################################################################
# I'm sure this could be improved...but real docs should be written
# anyway so there's no need to go overboard here.  That's my excuse,
# in any event.
#
sub display_usage {
  local (@formatlist, $item);

  print "#" x 72, "\n";
  print <<'EOF';
Usage: texit <options> texfile[.tex]
Where <options> may be:
  -<format>       process `texfile' using the named format.
  -once           process document once and exit.  Returns 0
                  if the document seems ok.
  -verbose        display additional status messages.
  -initfile init  load the file "init" during initialization.
  -tex            force an initial TeX run.
  -bibtex         start by running BibTeX (instead of TeX).
  -makeindex      start by running MakeIndex.

TeXit tries to figure out the format of your document, so you don't
have to specify it.  Of course, you have to specify it if TeXit guesses
wrong.  Better yet, stick in an Emacs-style local var section to define
the format or a "% Format: format" line at the top of your document.

You only have to specify enough of the option string to make it
unambiguous (e.g. "-v" is good enough for -verbose).

EOF

  foreach $item (keys %formats) {
    unshift(@formatlist, $item) if !defined($casefoo{$item});
  }

  &print_long_list("TeXit recognizes the following format names:",
                   sort(@formatlist));
  print "#" x 72, "\n";
  exit 5;
}
